@@ -1,5 +1,5 @@
 /***************************************************************************
- *   Copyright (C) 2007-2009 by Peter Semiletov                            *
+ *   Copyright (C) 2007-2010 by Peter Semiletov                            *
  *   peter.semiletov@gmail.com                                             *
  *                                                                         *
  *   This program is free software; you can redistribute it and/or modify  *
@@ -28,6 +28,7 @@
 ** Foundation and appearing in the files LICENSE.GPL2 and LICENSE.GPL3
 ** included in the packaging of this file.
 */
+
 /*
 code from qPEditor:
 
@@ -35,8 +36,16 @@ code from qPEditor:
 *   mik-protasov@anyqsoft.com
 
 */
+
 /*
 Diego Iastrubni <elcuco@kde.org> //some GPL'ed code from new-editor-diego-3, found on qtcentre forum
+*/
+
+/*
+code from qwriter:
+
+ *   Copyright (C) 2009 by Gancov Kostya                                   *
+ *   kossne@mail.ru                                                        *
 
 */
 
@@ -92,9 +101,8 @@ void CDocument::update_title (bool fullname)
 
 void CDocument::reload (const QString &enc)
 {
-  textEdit->setPlainText (qstring_load (file_name, enc.toUtf8().data()));
-  textEdit->document()->setModified (false);
-  holder->log->log (tr ("%1 is reloaded").arg (file_name));
+  if (file_exists (file_name))
+      open_file (file_name, enc);
 }
 
 
@@ -149,6 +157,14 @@ bool CDocument::save_with_name (const QString &fileName, const QString &codec)
       tio->charset = codec;
 
   tio->data = textEdit->toPlainText();
+  
+  if (eol_changed && (new_eol != def_eol))
+      tio->data.replace ("\n", new_eol);
+   
+#ifdef Q_WS_WIN   
+  if (! eol_changed)
+      tio->data.replace ("\n", def_eol);
+#endif
 
   if (! tio->save (fileName))
      {
@@ -177,19 +193,35 @@ bool CDocument::save_with_name (const QString &fileName, const QString &codec)
 CDocument::CDocument (QObject *parent): QObject (parent)
 {
   QString fname = tr ("new[%1]").arg (QTime::currentTime().toString ("hh-mm-ss"));
-
-  markup_mode = "HTML";
+  
+  fnameswoexts.insert ("configure", "sh");
+  fnameswoexts.insert ("install-sh", "sh");
+  
+  markup_mode = "HML";
   file_name = fname;
   cursor_xy_visible = true;
   charset = "UTF-8";
   text_to_search = "";
   position = 0;
+  eol_changed = false;
+  
+#ifdef Q_WS_WIN
+
+  def_eol = "\r\n";
+
+#else
+
+  def_eol = "\n";
+
+#endif
+  
+  new_eol = def_eol;
 }
 
 
 CDocument::~CDocument()
 {
-  if (file_name == holder->fname_crapbook && textEdit->document()->isModified())
+  if (file_name.startsWith (holder->dir_config) && textEdit->document()->isModified()) 
      save_with_name_plain (file_name);
   else
   if (textEdit->document()->isModified() && file_exists (file_name))
@@ -204,18 +236,23 @@ CDocument::~CDocument()
      }
 
   if (file_name != holder->fname_crapbook)
+  if (! file_name.startsWith (holder->dir_config))
      {
       holder->add_to_recent (this);
       holder->update_recent_menu();
      }
 
+  if (file_name.startsWith (holder->todo.dir_days))
+      holder->todo.load_dayfile();
+
+
   int i = holder->tab_widget->indexOf (tab_page);
   if (i != -1)
      holder->tab_widget->removeTab (i);
 }
 
 
-void CDocument::create_new (void)
+void CDocument::create_new()
 {
   textEdit = new CTEAEdit;
 
@@ -234,7 +271,7 @@ void CDocument::create_new (void)
 }
 
 
-int CDocument::get_tab_idx (void)
+int CDocument::get_tab_idx()
 {
   return holder->tab_widget->indexOf (tab_page);
 }
@@ -264,7 +301,7 @@ document_holder::~document_holder()
 }
 
 
-CDocument* document_holder::create_new (void)
+CDocument* document_holder::create_new()
 {
   CDocument *doc = new CDocument;
 
@@ -279,6 +316,7 @@ CDocument* document_holder::create_new (void)
 
   doc->update_title();
   doc->update_status();
+  
   return doc;
 }
 
@@ -297,11 +335,11 @@ CDocument* document_holder::open_file (const QString &fileName, const QString &c
      }   
    
   foreach (CDocument *d, list)  
-         if (d->file_name == fileName)
-          {
-           tab_widget->setCurrentIndex (tab_widget->indexOf (d->tab_page));
-           return d;
-          }
+          if (d->file_name == fileName)
+             {
+              tab_widget->setCurrentIndex (tab_widget->indexOf (d->tab_page));
+              return d;
+             }
    
   CDocument *doc = create_new();
   doc->open_file (fileName, codec);
@@ -309,6 +347,7 @@ CDocument* document_holder::open_file (const QString &fileName, const QString &c
   doc->update_title();
 
   tab_widget->setCurrentIndex (tab_widget->indexOf (doc->tab_page));
+  
   return doc;
 }
 
@@ -324,7 +363,7 @@ void document_holder::close_by_idx (int i)
 }
 
 
-void document_holder::close_current (void)
+void document_holder::close_current()
 {
   int i = tab_widget->currentIndex();
   if (i < 0)
@@ -411,7 +450,13 @@ void CDocument::set_hl (bool mode_auto, const QString &theext)
      ext = file_get_ext (file_name);
   else
       ext = theext;
-      
+
+  if (ext.isNull() || ext.isEmpty())
+    {
+     QFileInfo fi (file_name);
+     ext = fnameswoexts[fi.fileName()];
+    }
+
   if (ext.isNull() || ext.isEmpty())
      return;
 
@@ -425,7 +470,7 @@ void CDocument::set_hl (bool mode_auto, const QString &theext)
   
   highlighter = new CSyntaxHighlighter (textEdit->document(), this);
   highlighter->load_from_xml (fname);
-    
+      
   if (textEdit->use_hl_wrap)
      {
       if (highlighter->wrap)
@@ -453,6 +498,7 @@ void document_holder::apply_settings_single (CDocument *d)
   f.fromString (settings->value ("editor_font_name", "Monospace").toString());
   f.setPointSize (settings->value ("editor_font_size", "12").toInt());
   d->textEdit->setFont (f);
+  d->textEdit->lineNumberArea->setFont (f);
 
   d->textEdit->setCenterOnScroll (settings->value ("center_on_scroll", true).toBool());
   
@@ -469,11 +515,11 @@ void document_holder::apply_settings_single (CDocument *d)
   d->textEdit->setTabStopWidth (d->textEdit->spaces_instead_of_tabs * d->textEdit->brace_width);
   
   d->textEdit->setup_brace_width();
-  
+
+  d->textEdit->set_show_linenums (settings->value ("show_linenums", false).toBool());
   d->textEdit->set_show_margin (settings->value ("show_margin", false).toBool());
   d->textEdit->set_margin_pos (settings->value ("margin_pos", 72).toInt());
 
-
   d->textEdit->highlightCurrentLine = settings->value ("additional_hl", false).toBool();
   d->textEdit->hl_brackets = settings->value ("hl_brackets", false).toBool();
 
@@ -484,10 +530,15 @@ void document_holder::apply_settings_single (CDocument *d)
   d->textEdit->auto_indent = settings->value ("auto_indent", false).toBool();
 
   QString text_color = hash_get_val (palette, "text", "black");
+  d->textEdit->text_color = QColor (text_color);
+
   QString back_color = hash_get_val (palette, "background", "white");
   QString sel_back_color = hash_get_val (palette, "sel-background", "black");
   QString sel_text_color = hash_get_val (palette, "sel-text", "white");
 
+  d->textEdit->margin_color = QColor (hash_get_val (palette, "margin_color", text_color));
+  d->textEdit->linenums_bg = QColor (hash_get_val (palette, "linenums_bg", back_color));
+
   QString sheet = QString ("QPlainTextEdit { color: %1 ; background-color: %2 ; selection-color: %3; selection-background-color: %4;}").arg (
                            text_color).arg (
                            back_color).arg (
@@ -500,7 +551,7 @@ void document_holder::apply_settings_single (CDocument *d)
 }
 
 
-void document_holder::apply_settings (void)
+void document_holder::apply_settings()
 {
   foreach (CDocument *d, list)
           apply_settings_single (d);
@@ -524,14 +575,14 @@ void document_holder::add_to_recent (CDocument *d)
 }
 
 
-void document_holder::update_recent_menu ()
+void document_holder::update_recent_menu()
 {
   recent_menu->clear();
   create_menu_from_list (this, recent_menu, recent_files, SLOT(open_recent()));
 }
 
 
-void document_holder::open_recent ()
+void document_holder::open_recent()
 {
   QAction *Act = qobject_cast<QAction *>(sender());
 
@@ -564,7 +615,7 @@ QString CDocument::get_triplex()
   s += ",";
   s += charset;
   s += ",";
-  s += QString ("%1").arg (textEdit->textCursor().position());
+  s += QString::number (textEdit->textCursor().position()); 
   return s;
 }
 
@@ -602,8 +653,12 @@ void CTEAEdit::dragEnterEvent (QDragEnterEvent *event)
 void CTEAEdit::cb_cursorPositionChanged()
 {
   viewport()->update();
+  
   if (doc)
      doc->update_status();
+     
+  if (hl_brackets)   
+     braceHighlight();   
 }
 
 
@@ -642,6 +697,14 @@ void CTEAEdit::set_show_margin (bool enable)
 }
 
 
+void CTEAEdit::set_show_linenums (bool enable)
+{
+  draw_linenums = enable;
+  updateLineNumberAreaWidth (0);
+  emit repaint();
+}
+
+
 void CTEAEdit::set_margin_pos (int mp)
 {
   margin_pos = mp;
@@ -655,6 +718,9 @@ void CTEAEdit::paintEvent (QPaintEvent *event)
   if (draw_margin)
      {
       QPainter painter (viewport());
+      QPen pen = painter.pen();
+      pen.setColor (margin_color);
+      painter.setPen (pen);
       painter.drawLine (margin_x, 0, margin_x, viewport()->height());
       painter.end();
      }  
@@ -670,46 +736,6 @@ void CTEAEdit::paintEvent (QPaintEvent *event)
       painter.end();
      }
 
-   if (hl_brackets)
-      {
-       QTextCursor cur = textCursor();
-       int pos = cur.position();
-       QTextBlock block = document()->findBlock (pos);
-       QString tmp = block.text();
-
-       if (tmp.isNull() || tmp.isEmpty())
-          {
-           QPlainTextEdit::paintEvent (event);
-           return;
-          }
-
-       QChar charCC = tmp.at (pos - block.position());
-
-       int cc = tmp.at (pos - block.position()).toLatin1();
-       if (exists (pos))
-          switch (cc)
-                {
-                 case '{':
-                          paintBracketForward (cur, pos, QChar ('{'));
-                          break;
-                 case '}':
-                          paintBracketBack (cur, pos, QChar('{'));
-                          break;
-                 case '(':
-                          paintBracketForward (cur, pos, QChar('('));
-                          break;
-                 case ')':
-                          paintBracketBack (cur, pos, QChar('('));
-                          break;
-                 case '[':
-                          paintBracketForward (cur, pos , QChar('['));
-                          break;
-                 case ']':
-                          paintBracketBack (cur, pos , QChar('['));
-                          break;
-                }
-      }
-
   QPlainTextEdit::paintEvent (event);
 }
 
@@ -717,7 +743,7 @@ void CTEAEdit::paintEvent (QPaintEvent *event)
 void CTEAEdit::setup_brace_width()
 {
   QFontMetrics *fm = new QFontMetrics (currentCharFormat().font());
-  brace_width = fm->width ("a");
+  brace_width = fm->width ("9");
 }
 
 
@@ -726,6 +752,11 @@ CTEAEdit::CTEAEdit (QWidget *parent): QPlainTextEdit (parent)
   highlightCurrentLine = false;
   setup_brace_width();
   
+  lineNumberArea = new LineNumberArea (this);
+  connect(this, SIGNAL(blockCountChanged(int)), this, SLOT(updateLineNumberAreaWidth(int)));
+  connect(this, SIGNAL(updateRequest(const QRect &, int)), this, SLOT(updateLineNumberArea(const QRect &, int)));
+  updateLineNumberAreaWidth (0);
+
   margin_pos = 72;
   margin_x = brace_width * margin_pos;
   draw_margin = false;
@@ -791,8 +822,8 @@ QStringList CDocument::get_words()
         }  
            
      if (! stext.isNull() || ! stext.isEmpty())
-         result.append (stext);     
-     }
+        result.append (stext);     
+    }
   while (cr.movePosition (QTextCursor::NextWord));
 
   return result;
@@ -892,19 +923,21 @@ QTextCharFormat tformat_from_style (const QString &fontstyle, const QString &col
 
 void CSyntaxHighlighter::load_from_xml (const QString &fname)
 {
+  exts = "default";   
+  langs = "default";   
+  
   QDomDocument doc ("mydocument");
 
   if (! doc.setContent (qstring_load (fname)))
-     {
-      qDebug() << "cannot open";
       return;
-     }
 
-  QTextCharFormat tformat;
   QDomElement docElem = doc.documentElement();
 
-  QDomNode n = docElem.firstChild();
+  exts = docElem.attribute ("exts", "default");   
+  langs = docElem.attribute ("langs", "default");   
 
+  QDomNode n = docElem.firstChild();
+  
   while (! n.isNull())
         {
          QDomElement e = n.toElement();
@@ -969,286 +1002,24 @@ void CSyntaxHighlighter::load_from_xml (const QString &fname)
 
              if (e.attribute("type") == "mcomment-end")
                 commentEndExpression = QRegExp (e.text().trimmed());
+                
+                
+             if (e.attribute("type") == "comment")
+                {
+                 if (e.attribute ("name") == "cm_mult")
+                    cm_mult = e.text().trimmed();
+
+                 if (e.attribute ("name") == "cm_single")
+                    cm_single = e.text().trimmed();
+                }
+   
+                
             }
 
          n = n.nextSibling();
         }
 }
 
-//
-void CTEAEdit::paintBracketForward (QTextCursor &cur, int pos, const QChar &begin)
-{
-  QBrush brush (brackets_color);
-  QPainter painter (viewport());
-
-  QRect r = cursorRect(cur);
-  r.setWidth (brace_width);
-  painter.fillRect (r, brush);
-
-  int ends = ending (pos, begin);
-  if (ends != -1)
-     {
-      cur.movePosition (QTextCursor::Right, QTextCursor::MoveAnchor, ends - pos);
-
-      r = cursorRect (cur);
-      r.setWidth (brace_width);
-      painter.fillRect(r, brush);
-     }
-}
-
-
-void CTEAEdit::paintBracketBack (QTextCursor &cur, int pos, const QChar &begin)
-{
-  QBrush brush (brackets_color);
-  QPainter painter (viewport());
-
-  QRect r = cursorRect();
-  r.setWidth (brace_width);
-
-  painter.fillRect (r, brush);
-
-  int forwards = forwarding (pos, begin);
-  if (forwards != -1)
-     {
-      cur.movePosition (QTextCursor::Left, QTextCursor::MoveAnchor, pos - forwards);
-
-      r = cursorRect (cur);
-      r.setWidth (brace_width);
-      painter.fillRect (r, brush);
-     }
-}
-
-
-bool CTEAEdit::exists (int pos)
-{
-  TextType = Text_OFF;
-  bool ret = false;
-  int cc;
-  
-  QString tmp = document()->toPlainText();
-
-  for (int i = 0; i < pos; i++)
-      {
-       cc = tmp.at(i).toLatin1();
-       switch (TextType)
-              {
-               case Text_OFF:
-                             if (cc == '\'')
-                                TextType = Text_CHAR;
-                             if (cc == '"')
-                                TextType = Text_STRING;
-                             if (i > 0 && cc == '*' && tmp.at (i - 1).toLatin1() == '/')
-                                TextType = Text_MULTILINECOMMENT;
-                             if (i > 0 && cc == '/' && tmp.at (i - 1).toLatin1() == '/')
-                                TextType = Text_SINGLELINECOMMENT;
-                             break;
-
-               case Text_SINGLELINECOMMENT: if (cc == 0x0A)
-                                               TextType = Text_OFF;
-                                            break;
-
-               case Text_MULTILINECOMMENT: if (i > 0 && cc == '/' && tmp.at (i - 1).toLatin1() == '*')
-                                              TextType = Text_OFF;
-                                           break;
-
-               case Text_CHAR: if (cc == '\'')
-                                  TextType = Text_OFF;
-                               break;
-
-               case Text_STRING: if (cc == '"')
-                                     TextType = Text_OFF;
-                                 break;
-              }
-       }
-
-  if (TextType == Text_OFF)
-      ret = true;
-
-   return ret;
-}
-
-
-int CTEAEdit::ending (int pos, const QChar &ch)
-{
-  int count = 0;
-  int current = -1;
-  QChar begin = ch;
-  QChar end;
-
-  switch (begin.toLatin1())
-         {
-          case '{': end = QChar('}'); break;
-          case '(': end = QChar(')'); break;
-          case '[': end = QChar(']'); break;
-         }
-
-  TextType = Text_OFF;
-  int cc;
-  QString tmp = document()->toPlainText();
-  int tlen = tmp.length();
-
-  for (int i = pos; i < tlen; i++)
-      {
-       cc = tmp.at (i).toLatin1();
-       switch (TextType)
-              {
-               case Text_OFF:
-                             if (cc == '\'')
-                                TextType = Text_CHAR;
-
-                             if (cc == '"')
-                                TextType = Text_STRING;
-
-                             if (i > 0 && cc == '*' && tmp.at(i - 1).toLatin1() == '/')
-                                TextType = Text_MULTILINECOMMENT;
-
-                             if (i > 0 && cc == '/' && tmp.at( i - 1 ).toLatin1() == '/')
-                                TextType = Text_SINGLELINECOMMENT;
-
-                             if (cc == end.toLatin1())
-                                {
-                                 count--;
-                                 if (count == 0)
-                                    {
-                                     current = i;
-                                     break;
-                                    }
-                                 }
-
-                             if (cc == begin.toLatin1())
-                                 count++;
-                             break;
-
-               case Text_SINGLELINECOMMENT:
-                                           if (cc == 0x0A)
-                                              TextType = Text_OFF;
-                                           break;
-
-               case Text_MULTILINECOMMENT:
-                                          if (i > 0 && cc == '/' && tmp.at(i - 1).toLatin1() == '*')
-                                             TextType = Text_OFF;
-                                          break;
-
-               case Text_CHAR:
-                              if (cc == '\'')
-                                 TextType = Text_OFF;
-                              break;
-
-               case Text_STRING:
-                                if (cc == '"')
-                                   TextType = Text_OFF;
-                                break;
-              }
-
-       if (current != -1) break;
-      }
-
-  return current;
-}
-
-
-int CTEAEdit::forwarding (int pos, const QChar &ch)
-{
-  QList <QChar> lstCh;
-  QList <int> lstPos;
-  int count = 0;
-  int i;
-  int current = -1;
-  QChar begin = ch;
-  QChar end;
-
-  switch (begin.toLatin1())
-         {
-          case '{':
-                   end = QChar ('}');
-                   break;
-
-          case '(':
-                   end = QChar (')');
-                   break;
-
-          case '[':
-                   end = QChar( ']' );
-                   break;
-         }
-
-  TextType = Text_OFF;
-  int cc;
-  QString tmp = document()->toPlainText();
-
-  for (i = 0; i < pos + 1; i++)
-      {
-       cc = tmp.at( i ).toLatin1();
-       switch (TextType)
-              {
-               case Text_OFF:
-                             if ( cc == '\'' )
-                                TextType = Text_CHAR;
-
-                             if ( cc == '"' )
-                                TextType = Text_STRING;
-
-                             if (i > 0 && cc == '*' && tmp.at( i - 1 ).toLatin1() == '/')
-                                TextType = Text_MULTILINECOMMENT;
-
-                             if (i > 0 && cc == '/' && tmp.at( i - 1 ).toLatin1 () == '/')
-                                TextType = Text_SINGLELINECOMMENT;
-
-                             if (cc == end.toLatin1())
-                                {
-                                 lstCh.append(end);
-                                 lstPos.append(i);
-                                }
-
-                             if (cc == begin.toLatin1())
-                                {
-                                 lstCh.append(begin);
-                                 lstPos.append(i);
-                                }
-
-                             break;
-
-               case Text_SINGLELINECOMMENT:
-                                           if (cc == 0x0A)
-                                              TextType =  Text_OFF;
-                                           break;
-
-               case Text_MULTILINECOMMENT:
-                                          if (i > 0 && cc == '/' && tmp.at(i - 1).toLatin1() == '*')
-                                             TextType = Text_OFF;
-                                          break;
-
-               case Text_CHAR:
-                              if (cc == '\'')
-                                 TextType = Text_OFF;
-                              break;
-
-               case Text_STRING:
-                                if (cc == '"')
-                                   TextType = Text_OFF;
-                                break;
-              }
-      }
-
-  count = 0;
-  int index = lstPos.indexOf(pos);
-
-  for (i = index; i >= 0; i--)
-      {
-       if (lstCh.at(i) == end)
-           count++;
-       else count--;
-
-       if (count == 0)
-          {
-           current = lstPos.at(i);
-           break;
-          }
-      }
-
-  return current;
-}
-
 
 CTEAEdit::~CTEAEdit()
 {
@@ -1258,8 +1029,7 @@ CTEAEdit::~CTEAEdit()
 
 void document_holder::save_to_session (const QString &fileName)
 {
-  int c = list.size();
-  if (c < 0)
+  if (list.size() < 0)
      return;
 
   fname_current_session = fileName;
@@ -1334,35 +1104,7 @@ bool CDocument::get_has_selection() const
 
 void CDocument::insert_image (const QString &full_path)
 {
-  if (! is_image (full_path))
-     return;     
-
-  QFileInfo inf (file_name);
-  QDir dir (inf.absolutePath());
-
-  QImage img;
-  img.load (full_path);
-  QString result;
-
-  if (markup_mode == "HTML")
-     result = QString ("<img src=\"%1\" border=\"0\" alt=\"\" width=\"%2\" height=\"%3\">").arg (
-                       dir.relativeFilePath (full_path)).arg (img.width()).arg (img.height());
-  else
-  if (markup_mode == "XHTML")
-     result = QString ("<img src=\"%1\" border=\"0\" alt=\"\" width=\"%2\" height=\"%3\" />").arg (
-                       dir.relativeFilePath (full_path)).arg (img.width()).arg (img.height());
-  else
-  if (markup_mode == "Docbook")
-     result = QString ("<mediaobject><imageobject>\n<imagedata fileref=\"%1\"/>\n</imageobject></mediaobject>").arg (
-                        dir.relativeFilePath (full_path)) ;
-  else
-  if (markup_mode == "LaTeX")
-      result = QString ("\\includegraphics{%1}").arg (dir.relativeFilePath (full_path));
-  else
-  if (markup_mode == "Lout")
-      result = QString ("@IncludeGraphic {%1}").arg (dir.relativeFilePath (full_path));
-
-   textEdit->textCursor().insertText (result);
+  textEdit->textCursor().insertText (get_insert_image (file_name, full_path, markup_mode));
 }
 
 
@@ -1495,3 +1237,220 @@ void CTEAEdit::keyPressEvent (QKeyEvent *event)
   
   QPlainTextEdit::keyPressEvent (event);
 }
+
+
+int CTEAEdit::lineNumberAreaWidth()
+{
+  if (! draw_linenums)
+     return 0; 
+
+  int digits = 1;
+  int max = qMax (1, blockCount());
+  
+  while (max >= 10) 
+        {
+         max /= 10;
+         ++digits;
+        }
+
+  int space = 18 + brace_width * digits;
+
+  return space;
+}
+
+
+void CTEAEdit::updateLineNumberAreaWidth (int newBlockCount)
+{
+  setViewportMargins (lineNumberAreaWidth(), 0, 0, 0);
+}
+
+
+void CTEAEdit::updateLineNumberArea (const QRect &rect, int dy)
+{
+  if (dy)
+     lineNumberArea->scroll (0, dy);
+  else
+     lineNumberArea->update (0, rect.y(), lineNumberArea->width(), rect.height());
+
+  if (rect.contains (viewport()->rect()))
+     updateLineNumberAreaWidth (0);
+}
+
+
+void CTEAEdit::lineNumberAreaPaintEvent (QPaintEvent *event)
+{
+  if (! draw_linenums)
+     return;
+
+  QPainter painter (lineNumberArea);
+  painter.fillRect (event->rect(), linenums_bg);
+
+  QTextBlock block = firstVisibleBlock();
+  int blockNumber = block.blockNumber();
+  int top = (int) blockBoundingGeometry (block).translated (contentOffset()).top();
+  int bottom = top + (int) blockBoundingRect (block).height();
+
+  while (block.isValid() && top <= event->rect().bottom()) 
+        {
+         if (block.isVisible() && bottom >= event->rect().top()) 
+            {
+             QString number = QString::number (blockNumber + 1) + " ";
+             painter.setPen (text_color);
+             painter.drawText (0, top, lineNumberArea->width(), fontMetrics().height(), Qt::AlignRight, number);
+             /*int index = bookMark.indexOf(number.toInt());
+ 
+             if (index != -1) 
+                {
+                //painter.drawText(0, top, 30, fontMetrics().height(), Qt::AlignCenter, "+");
+                 painter.setBrush(QBrush(Qt::yellow, Qt::SolidPattern));
+                 painter.drawEllipse(5, top + (fontMetrics().height()/4), fontMetrics().height()/2, fontMetrics().height()/2);
+                }*/
+              }
+
+         block = block.next();
+         top = bottom;
+         bottom = top + (int) blockBoundingRect(block).height();
+         ++blockNumber;
+        }
+}
+
+
+void CTEAEdit::resizeEvent(QResizeEvent *e)
+{
+  QPlainTextEdit::resizeEvent (e);
+  QRect cr = contentsRect();
+  lineNumberArea->setGeometry (QRect(cr.left(), cr.top(), lineNumberAreaWidth(), cr.height()));
+}
+
+
+void CTEAEdit::braceHighlight()
+{
+  extraSelections.clear();
+  setExtraSelections(extraSelections);
+  selection.format.setBackground(brackets_color);
+
+  QTextDocument *doc = document();
+  QTextCursor cursor = textCursor();
+  QTextCursor beforeCursor = cursor;
+
+  cursor.movePosition(QTextCursor::NextCharacter, QTextCursor::KeepAnchor);
+  QString brace = cursor.selectedText();
+
+  beforeCursor.movePosition(QTextCursor::PreviousCharacter, QTextCursor::KeepAnchor);
+  QString beforeBrace = beforeCursor.selectedText();
+
+  if ((brace != "{") && 
+      (brace != "}") &&
+      (brace != "[") &&
+      (brace != "]") && 
+      (brace != "(") && 
+      (brace != ")") && 
+      (brace != "<") && 
+      (brace != ">")) 
+      {
+       if ((beforeBrace == "{") 
+           || (beforeBrace == "}") 
+           || (beforeBrace == "[")
+           || (beforeBrace == "]")
+           || (beforeBrace == "(")
+           || (beforeBrace == ")")
+           || (beforeBrace == "<")
+           || (beforeBrace == ">")) 
+          {
+           cursor = beforeCursor;
+           brace = cursor.selectedText();
+          }
+        else 
+            return;
+       }
+
+  QString openBrace;
+  QString closeBrace;
+
+  if ((brace == "{") || (brace == "}")) 
+     {
+      openBrace = "{";
+      closeBrace = "}";
+     }
+  else
+  if ((brace == "[") || (brace == "]")) 
+     {
+      openBrace = "[";
+      closeBrace = "]";
+     }
+  else
+  if ((brace == "(") || (brace == ")")) 
+     {
+      openBrace = "(";
+      closeBrace = ")";
+     }
+  else
+  if ((brace == "<") || (brace == ">")) 
+     {
+      openBrace = "<";
+      closeBrace = ">";
+     }
+
+
+  if (brace == openBrace) 
+     {
+      QTextCursor cursor1 = doc->find (closeBrace, cursor);
+      QTextCursor cursor2 = doc->find (openBrace, cursor);
+      if (cursor2.isNull()) 
+         {
+          selection.cursor = cursor;
+          extraSelections.append (selection);
+          selection.cursor = cursor1;
+          extraSelections.append (selection);
+          setExtraSelections (extraSelections);
+         } 
+      else 
+          {
+           while (cursor1.position() > cursor2.position()) 
+                 {
+                  cursor1 = doc->find (closeBrace, cursor1);
+                  cursor2 = doc->find (openBrace, cursor2);
+                  if (cursor2.isNull()) 
+                      break;
+                 }
+                
+           selection.cursor = cursor;
+           extraSelections.append (selection);
+           selection.cursor = cursor1;
+           extraSelections.append (selection);
+           setExtraSelections (extraSelections);
+          }
+       } 
+   else
+       {
+        if (brace == closeBrace) 
+           {
+            QTextCursor cursor1 = doc->find (openBrace, cursor, QTextDocument::FindBackward);
+            QTextCursor cursor2 = doc->find (closeBrace, cursor, QTextDocument::FindBackward);
+            if (cursor2.isNull()) 
+               {
+                selection.cursor = cursor;
+                extraSelections.append(selection);
+                selection.cursor = cursor1;
+                extraSelections.append (selection);
+                setExtraSelections (extraSelections);
+               }
+            else 
+                {
+                 while (cursor1.position() < cursor2.position()) 
+                       {
+                        cursor1 = doc->find (openBrace, cursor1, QTextDocument::FindBackward);
+                        cursor2 = doc->find (closeBrace, cursor2, QTextDocument::FindBackward);
+                        if (cursor2.isNull()) 
+                           break;
+                       }
+                       
+                 selection.cursor = cursor;
+                 extraSelections.append (selection);
+                 selection.cursor = cursor1;
+                 extraSelections.append (selection);
+                 setExtraSelections (extraSelections);
+                }
+             }
+     }
+}
